/* (c) 2004 Matthew Garrett, almost entirely based on read-edid */
/* (c) 2000,2001 John Fremlin */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <sys/types.h>
#include <sys/io.h>
#include <pci/pci.h>

#include "lrmi.h"

typedef u_int32_t uint32;

typedef struct LRMI_regs reg_frame;
typedef byte* real_ptr;

static struct pci_access *pacc;

#define access_register(reg_frame,reg) (reg_frame . reg)
#define access_ptr_register(reg_frame,reg) (reg_frame -> reg)
#define access_seg_register(reg_frame,es) reg_frame.es
#define real_mode_int(interrupt,reg_frame_ptr) !LRMI_int(interrupt,reg_frame_ptr)
#define real_mode_call(reg_frame_ptr) !LRMI_call(reg_frame_ptr)
#define alloc_real(size) LRMI_alloc_real(size)
#define free_real(block) LRMI_free_real(block)
#define dosmemput(buffer,length,offset) memcpy(offset,buffer,length)

int do_post(void);

int
main( int argc, char** argv )
{
	int error=0;
	
	if( !LRMI_init() )
	{
		fprintf( stderr, "%s: error initialising realmode interface\n", argv[0] );
		fprintf( stderr, "%s: do you have full superuser (root) permissions?\n", argv[0] );
		return 10;
	}
	
	pacc = pci_alloc();
	pacc->numeric_ids = 1;
	pci_init(pacc);


	ioperm(0, 0x400 , 1);
	iopl(3);
	
	if ( argc != 1) {
		fprintf( stderr, "%s: syntax %s\n",
			 argv[0], argv[0]);
		return 3;
	} else {
		do_post();
	}
	
	return error;
}

int do_real_post(unsigned pci_device,reg_frame* regs)
{
  int error = 0;

  access_ptr_register(regs,eax) = pci_device;
  
  access_ptr_register(regs,cs) = 0xc000;
  access_ptr_register(regs,ip) = 0x0003;

  /* This is all heavily cargo culted */
  access_ptr_register(regs,edx) = 0x80;
  access_ptr_register(regs,ss) = 0x0030;
  access_ptr_register(regs,ds) = 0x0040;
  access_ptr_register(regs,sp) = 0xfffe;

  if(real_mode_call(regs) )
    {
      fprintf( stderr, "Error: something went wrong performing real mode call\n" );
      error = 1;
    }

  return error;
}
  
int
do_post()
{
  reg_frame regs;
  struct pci_dev *p;
  unsigned int c;
  unsigned int pci_id;

  memset(&regs, 0, sizeof(regs));
  pci_scan_bus(pacc);

  for(p=pacc->devices; p; p=p->next) {
	  c=pci_read_word(p, PCI_CLASS_DEVICE);
	  if (c==0x300) {
		  pci_id = (p->bus <<8) + (p->dev << 3) + (p->func & 0x7);
		  do_real_post(pci_id,&regs);
	  }		  
  }
  return 0;
}
